use std::rc::{Rc, Weak};
use std::cell::RefCell;
use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;
use r9_macro_derive::Singleton;
use r9::Queue;
use f9::app::App;
use crate::listener::WsListener;
use crate::upstream::UpStreamMgr;
use bytes::BytesMut;
use r9::xtimer::TimerList;
use std::sync::OnceLock;

pub struct DownStreamPack {
    pub holder: Option<Rc::<RefCell::<DownStreamPack>>>,
    pub socket_handle: u16,
    pub data: bytes::BytesMut,
    pub entry: Rc::<RefCell::<r9::ListHead<DownStreamPack>>>,
}

unsafe impl Send for DownStreamPack {}
unsafe impl Sync for DownStreamPack {}

pub struct UserApp {
    net_msg_queue: Rc::<RefCell::<r9::Queue<DownStreamPack>>>,
}
unsafe impl Send for UserApp {}
unsafe impl Sync for UserApp {}

impl f9::app::UserApp for UserApp {

    fn get_pkg_name(&self) -> String {
        return "statserver".to_string();
    }

    fn init(&self) {
        WsListener::instance().borrow_mut().init();
        UpStreamMgr::instance().borrow_mut().init();
    }

    fn update(&self) {
        self.net_msg_queue.borrow_mut().fetch();
        let work_list = self.net_msg_queue.borrow().work_list.lock().unwrap().clone();
        while !work_list.borrow().empty() {
            let node = &work_list.borrow().first_entry();
            node.upgrade().unwrap().borrow_mut().entry.borrow_mut().del_init();
            WsListener::instance().borrow().add_down_stream_pack(node.clone());
        }
        WsListener::instance().borrow().update();
    }

    fn uninit(&self) {
        WsListener::instance().borrow_mut().uninit();
    }

    fn get_http_listen_port(&self) -> i32 {
        return 8080;
    }

}

static UserAppInstance: OnceLock<UserApp> = OnceLock::new();

impl UserApp {

    pub fn instance() -> &'static UserApp {
        return UserAppInstance.get_or_init(|| UserApp::new());
    }

    pub fn add_down_stream_pack(&self, socket_handle: u16, data: bytes::BytesMut) {
        println!("add_down_stream_pack1 len:{0}", data.len());
        let node = Rc::new(RefCell::new(DownStreamPack{
            holder: Default::default(),
            socket_handle: socket_handle,
            data: data,
            entry: r9::ListHead::< crate::app::user_app::DownStreamPack>::new_node(Default::default()),
        }));
        node.borrow_mut().holder = Some(node.clone());
        node.borrow_mut().entry = r9::ListHead::< crate::app::user_app::DownStreamPack>::new_node(Rc::downgrade(&node));
        self.net_msg_queue.borrow_mut().push(&node.borrow_mut().entry);
    }

    pub fn new() -> Self {
        let p = UserApp{
            net_msg_queue: r9::Queue::<DownStreamPack>::new(),
        };
        return p;
    }

}
