use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::sync::{Arc, Mutex};
use tokio::net::{TcpStream, ToSocketAddrs};
use tokio::io::AsyncWriteExt;
use tokio::runtime::Runtime;
use tokio::io::Interest;
use std::error::Error;
use std::io;

use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;
use r9_macro_derive::Singleton;


#[derive(SharedFromSelf)]
#[derive(Singleton)]
pub struct UpStreamMgr {
    _self_wp: Weak::<RefCell::<Self>>,
}

async fn run_app(send_queue: Arc::<Vec<Arc::<dyn ::protobuf::Message>>>) -> Result<(), Box<dyn Error>> {
    let mut stream = TcpStream::connect("192.168.100.39:7617").await?;

    loop {
        let ready = stream.ready(Interest::READABLE | Interest::WRITABLE).await?;

        if ready.is_readable() {
            let mut data = vec![0; 1024];
            // Try to read data, this may still fail with `WouldBlock`
            // if the readiness event is a false positive.
            match stream.try_read(&mut data) {
                Ok(n) => {
                    println!("read {} bytes", n);
                }
                Err(ref e) if e.kind() == io::ErrorKind::WouldBlock => {
                    continue;
                }
                Err(e) => {
                    //return Err(e.into());
                }
            }

        }

        if ready.is_writable() {
            // Try to write data, this may still fail with `WouldBlock`
            // if the readiness event is a false positive.
            match stream.try_write(b"hello world") {
                Ok(n) => {
                    println!("write {} bytes", n);
                }
                Err(ref e) if e.kind() == io::ErrorKind::WouldBlock => {
                    continue
                }
                Err(e) => {
                    //return Err(e.into());
                }
            }
        }
    }

    Ok(())
}

impl UpStreamMgr {

    pub fn new() -> Self {
        let p = UpStreamMgr{
            _self_wp: Default::default(),
        };
        return p;
    }

    pub fn init(&mut self) {
        let send_queue: Arc::<Vec<Arc::<dyn ::protobuf::Message>>> =
            Default::default();
        //let msg : Arc::<dyn ::protobuf::Message>>> = Arc::new();
        std::thread::spawn(move || {
            let rt = tokio::runtime::Builder::new_multi_thread()
                .worker_threads(4)
                .enable_all()
                .thread_name("my-custom-name")
                .thread_stack_size(3 * 1024 * 1024)
                .build()
                .unwrap();
            rt.block_on(async {
                run_app(send_queue.clone()).await;
            });
        });

    }

}
